# __all__ = ('dfchk')
import pandas as pd
import numpy as np
import os
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
from sklearn.model_selection import PredefinedSplit
# from sklearn.model_selection import GridSearchCV
# from sklearn.model_selection import StratifiedKFold
# from sklearn.metrics import precision_recall_fscore_support
# from sklearn.metrics import roc_auc_score

#%%
# Standard quick checks
def dfchk(dframe, chkNull=True, info=True, head=True, shape=True, desc=False, valCnt = False): 
  """
  some basic and common checks on dataframes. 
  Args:
      dframe (Pandas.DataFrame): pandas dataframe
      chkNull (bool): check Null option, defaults to True
      info (bool): info option, defaults to True
      head (bool): head option, defaults to True
      shape (bool): shape option, defaults to True
      desc (bool): describe option, defaults to False
      valCnt (bool): value_count option, defaults to False
  Return: None
  """
  cnt = 1
  print('\ndataframe Basic Check function -')
  
  if (chkNull):
    print(f'\n{cnt}: Null values:')
    cnt+=1
    print( dframe.isnull().sum().sort_values(ascending=False) )
  
  if (info):
    try:
      print(f'\n{cnt}: info(): ')
      cnt+=1
      print(dframe.info())
    except: pass
    
  if (desc):
    print(f'\n{cnt}: describe(): ')
    cnt+=1
    print(dframe.describe())
  
  if (head):
    print(f'\n{cnt}: head() -- ')
    cnt+=1
    print(dframe.head())

  if (shape): 
    print(f'\n{cnt}: shape: ')
    cnt+=1
    print(dframe.shape)

  if (valCnt):
    print('\nValue Counts for each feature -')
    for colname in dframe.columns :
      print(f'\n{cnt}: {colname} value_counts(): ')
      print(dframe[colname].value_counts())
      cnt +=1

# examples:
# dfchk(df, desc=True)

#%%
# from https://github.com/yuxiaohuang/teaching
def common_var_checker(df_train, df_val, df_test, target):
  """
  The common variables checker (Yuxiao Huang)

  Parameters
  ----------
  df_train : the dataframe of training data
  df_val : the dataframe of validation data
  df_test : the dataframe of test data
  target : the name of the target

  Returns
  ----------
  The dataframe of common variables between the training, validation and test data
  """
  
  # Get the dataframe of common variables between the training, validation and test data
  df_common_var = pd.DataFrame(np.intersect1d(np.intersect1d(df_train.columns, df_val.columns), np.union1d(df_test.columns, [target])), columns=['common var'])
              
  return df_common_var
  
#%%
# from https://github.com/yuxiaohuang/teaching
def id_checker(df, dtype='float'):
  """
  The identifier checker (Yuxiao Huang)

  Parameters
  ----------
  df : dataframe
  dtype : the data type identifiers cannot have, 'float' by default
          i.e., if a feature has this data type, it cannot be an identifier
  
  Returns
  ----------
  The dataframe of identifiers
  """
  
  # Get the dataframe of identifiers
  df_id = df[[var for var in df.columns
              # If the data type is not dtype
              if (df[var].dtype != dtype
                  # If the value is unique for each sample
                  and df[var].nunique(dropna=True) == df[var].notnull().sum())]]
  
  return df_id

#%%
# from https://github.com/yuxiaohuang/teaching
def datetime_transformer(df, datetime_vars):
  """
  The datetime transformer (Yuxiao Huang)

  Parameters
  ----------
  df : the dataframe
  datetime_vars : the datetime variables
  
  Returns
  ----------
  The dataframe where datetime_vars are transformed into the following 6 datetime types:
  year, month, day, hour, minute and second
  """
  
  # The dictionary with key as datetime type and value as datetime type operator
  dict_ = {'year'   : lambda x : x.dt.year,
            'month'  : lambda x : x.dt.month,
            'day'    : lambda x : x.dt.day,
            'hour'   : lambda x : x.dt.hour,
            'minute' : lambda x : x.dt.minute,
            'second' : lambda x : x.dt.second}
  
  # Make a copy of df
  df_datetime = df.copy(deep=True)
  
  # For each variable in datetime_vars
  for var in datetime_vars:
      # Cast the variable to datetime
      df_datetime[var] = pd.to_datetime(df_datetime[var])
      
      # For each item (datetime_type and datetime_type_operator) in dict_
      for datetime_type, datetime_type_operator in dict_.items():
          # Add a new variable to df_datetime where:
          # the variable's name is var + '_' + datetime_type
          # the variable's values are the ones obtained by datetime_type_operator
          df_datetime[var + '_' + datetime_type] = datetime_type_operator(df_datetime[var])
          
  # Remove datetime_vars from df_datetime
  df_datetime = df_datetime.drop(columns=datetime_vars)
              
  return df_datetime

#%%
# from https://github.com/yuxiaohuang/teaching
def nan_checker(df):
  """
  The NaN checker (Yuxiao Huang)

  Parameters
  ----------
  df : the dataframe
  
  Returns
  ----------
  The dataframe of variables with NaN, their proportion of NaN and data type
  """
  
  # Get the dataframe of variables with NaN, their proportion of NaN and data type
  df_nan = pd.DataFrame([[var, df[var].isna().sum() / df.shape[0], df[var].dtype]
                          for var in df.columns if df[var].isna().sum() > 0],
                        columns=['var', 'proportion', 'dtype'])
  
  # Sort df_nan in accending order of the proportion of NaN
  df_nan = df_nan.sort_values(by='proportion', ascending=False).reset_index(drop=True)
  
  return df_nan

#%%
# from https://github.com/yuxiaohuang/teaching
def cat_var_checker(df, dtype='object'):
  """
  The categorical variable checker (Yuxiao Huang)

  Parameters
  ----------
  df : the dataframe
  dtype : the data type categorical variables should have, 'object' by default
          i.e., if a variable has this data type, it should be a categorical variable
  
  Returns
  ----------
  The dataframe of categorical variables and their number of unique value
  """
  
  # Get the dataframe of categorical variables and their number of unique value
  df_cat = pd.DataFrame([[var, df[var].nunique(dropna=False)]
                          # If the data type is dtype
                          for var in df.columns if df[var].dtype == dtype],
                        columns=['var', 'nunique'])
  
  # Sort df_cat in accending order of the number of unique value
  df_cat = df_cat.sort_values(by='nunique', ascending=False).reset_index(drop=True)
  
  return df_cat

#%%
# from https://github.com/yuxiaohuang/teaching
def plot_scatter_x1_x2(X, y, xlim, xticks, xlabel, ylim, yticks, ylabel, title, dir_name, fig_name):
  """
  Plot the scatter plot between two features (Yuxiao Huang)
  
  Parameters
  ----------
  X : the feature matrix with two columns
  y : the target vector with two classes, 0 and 1
  xlim : x limits 
  xticks : x ticks
  xlabel : x label
  ylim : y limits 
  yticks : y ticks
  ylabel : y label
  title : the title of the scatter plot
  dir_name : the name of the directory
  fig_name : the name of the figure
  """
  
  # Make directory
  directory = os.path.dirname(dir_name)
  if not os.path.exists(directory):
      os.makedirs(directory)

  # The scatter plot
  plt.scatter(X[y == 1, 0],
              X[y == 1, 1],
              s=100,
              c='b',
              marker='x',
              label='1')
  plt.scatter(X[y == 0, 0],
              X[y == 0, 1],
              s=100,
              c='r',
              marker='s',
              label='0')

  # Set the x-axis
  plt.xlim(xlim)
  plt.xticks(xticks)
  plt.xlabel(xlabel)

  # Set the y-axis
  plt.ylim(ylim)
  plt.yticks(yticks)
  plt.ylabel(ylabel)

  # Set the title and legend
  plt.title(title)
  plt.legend(loc='best')
  
  # Save and show the figure
  plt.tight_layout()
  plt.savefig(dir_name + fig_name)
  plt.show()

#%%
# from https://github.com/yuxiaohuang/teaching
def plot_scatter_tsne(X, y, classes, labels, colors, markers, loc, dir_name, fig_name, random_seed):
  """
  Plot the scatter plot using TSNE (Yuxiao Huang)
  
  Parameters
  ----------
  X : the feature matrix
  y : the target vector
  classes : the classes in the target vector
  labels : the labels for different classes
  colors : the colors for different classes
  markers : the markers for different classes
  loc : the location of the legend
  dir_name : the name of the directory
  fig_name : the name of the figure
  random_seed : the random seed
  """
  
  # Make directory
  directory = os.path.dirname(dir_name)
  if not os.path.exists(directory):
      os.makedirs(directory)

  # Get the tsne transformed training feature matrix
  X_embedded = TSNE(n_components=2, random_state=random_seed).fit_transform(X)

  # Get the tsne dataframe
  tsne_df = pd.DataFrame(np.column_stack((X_embedded, y)), columns=['x1', 'x2', 'y'])

  # Get the data
  data = {}
  for class_ in classes:
      data_x1 = [tsne_df['x1'][i] for i in range(len(tsne_df['y'])) if tsne_df['y'][i] == class_]
      data_x2 = [tsne_df['x2'][i] for i in range(len(tsne_df['y'])) if tsne_df['y'][i] == class_]
      data[class_] = [data_x1, data_x2]
  
  # The scatter plot
  fig = plt.figure(figsize=(8, 6))
  
  for class_, label, color, marker in zip(classes, labels, colors, markers):
      data_x1, data_x2 = data[class_]
      plt.scatter(data_x1, data_x2, c=color, marker=marker, s=120, label=label)

  # Set x-axis
  plt.xlabel('x1')

  # Set y-axis
  plt.ylabel('x2')

  # Set legend
  plt.legend(loc=loc)

  # Save and show the figure
  plt.tight_layout()
  plt.savefig(dir_name + fig_name)
  plt.show()
    
#%%
# from https://github.com/yuxiaohuang/teaching
def separate_duplicate_original(X_aug_train, y_aug_train, minor_class):
  """
  Separate the duplicated class from the original class (Yuxiao Huang)

  Parameters
  ----------
  X_aug_train : The augmented feature matrix
  y_aug_train : The augmented target vector
  minor_class : The minority class
  
  Returns
  ----------
  The separated duplicated class and original class
  """

  # Make a copy of y_aug_train
  y_aug_dup_ori_train = np.array(y_aug_train)
  
  # For each sample in the augmented data
  for i in range(X_aug_train.shape[0]):
      # If the sample has the minor class
      if y_aug_dup_ori_train[i] == minor_class:
          # Flag variable, indicating whether a sample in the augmented data is the same as a sample in the original data
          same = False
          
          # For each sample in the original data
          for j in range(X_aug_train.shape[0]):
              if j == i:
                  continue

              # If the sample has the minor class
              if y_aug_dup_ori_train[j] == minor_class:
                  if len(np.setdiff1d(X_aug_train[i, :], X_aug_train[j, :])) == 0:
                      # The two samples are the same
                      same = True
                      break

          # If the two samples are different
          if same is False:
              y_aug_dup_ori_train[i] = 2
              
  return y_aug_dup_ori_train
  
#%%
# from https://github.com/yuxiaohuang/teaching
def separate_generate_original(X_aug_train, y_aug_train, X_train, y_train, minor_class):
  """
  Separate the generated class from the original class (Yuxiao Huang)

  Parameters
  ----------
  X_aug_train : The augmented feature matrix
  y_aug_train : The augmented target vector
  X_train : The original feature matrix
  y_train : The original target vector
  minor_class : The minority class
  
  Returns
  ----------
  The separated generated class and original class
  """
  
  # Make a copy of y_aug_train
  y_aug_gen_ori_train = np.array(y_aug_train)

  # For each sample in the augmented data
  for i in range(X_aug_train.shape[0]):
      # If the sample has the minor class
      if y_aug_gen_ori_train[i] == minor_class:
          # Flag variable, indicating whether a sample in the augmented data is the same as a sample in the original data
          same = False

          # For each sample in the original data
          for j in range(X_train.shape[0]):
              # If the sample has the minor class
              if y_train[j] == minor_class:
                  if len(np.setdiff1d(X_aug_train[i, :], X_train[j, :])) == 0:
                      # The two samples are the same
                      same = True
                      break

          # If the two samples are different
          if same is False:
              y_aug_gen_ori_train[i] = 2
              
  return y_aug_gen_ori_train

#%%
# from https://github.com/yuxiaohuang/teaching
def get_train_val_ps(X_train, y_train, X_val, y_val):
  """
  Get the:
  feature matrix and target velctor in the combined training and validation data
  target vector in the combined training and validation data
  PredefinedSplit
  
  Parameters
  ----------
  X_train : the feature matrix in the training data
  y_train : the target vector in the training data
  X_val : the feature matrix in the validation data
  y_val : the target vector in the validation data  

  Return
  ----------
  The feature matrix in the combined training and validation data
  The target vector in the combined training and validation data
  PredefinedSplit
  """  

  # Combine the feature matrix in the training and validation data
  X_train_val = np.vstack((X_train, X_val))

  # Combine the target vector in the training and validation data
  y_train_val = np.vstack((y_train.reshape(-1, 1), y_val.reshape(-1, 1))).reshape(-1)

  # Get the indices of training and validation data
  train_val_idxs = np.append(np.full(X_train.shape[0], -1), np.full(X_val.shape[0], 0))

  # The PredefinedSplit
  ps = PredefinedSplit(train_val_idxs)

  return X_train_val, y_train_val, ps

#%%
# NOT working
# from https://github.com/yuxiaohuang/teaching
# def training_valation_test(X_train, y_train, X_test, y_test, ps, abspath_curr, name):
#   """
#   Training, validation and test (Yuxiao Huang)
  
#   Parameters
#   ----------
#   X_train : the feature matrix in the training data
#   y_train : the target vector in the training data
#   X_test : the feature matrix in the test data
#   y_test : the target vector in the test data    
#   ps : the PredefinedSplit
#   abspath_curr : the absolute path of the current folder
#   name : the name of the cv_results folder

#   Return
#   ----------
#   The dataframe of [precision, recall, best_estimator]
#   """    
  
#   #************************************************************************************************
#   # Creating the directory for the cv results
#   directory = os.path.dirname(abspath_curr + name + '/')
#   if not os.path.exists(directory):
#       os.makedirs(directory)
      
#   #************************************************************************************************
#   # Training and validation

#   # The list of [best_score_, best_params_, best_estimator_] obtained by GridSearchCV
#   best_score_param_estimator_gs = []

#   for acronym in pipes.keys():
#       # GridSearchCV
#       gs = GridSearchCV(estimator=pipes[acronym],
#                         param_grid=param_grids[acronym],
#                         scoring='f1',
#                         n_jobs=2,
#                         cv=ps,
#                         return_train_score=True)

#       # Fit the pipeline
#       gs = gs.fit(X_train, y_train)

#       # Update best_score_param_estimator_gs
#       best_score_param_estimator_gs.append([gs.best_score_, gs.best_params_, gs.best_estimator_])

#       # Sort cv_results in ascending order of 'rank_test_score' and 'std_test_score'
#       cv_results = pd.DataFrame.from_dict(gs.cv_results_).sort_values(by=['rank_test_score', 'std_test_score'])

#       # Get the important columns in cv_results
#       important_columns = ['rank_test_score',
#                             'mean_test_score', 
#                             'std_test_score', 
#                             'mean_train_score', 
#                             'std_train_score',
#                             'mean_fit_time', 
#                             'std_fit_time',                        
#                             'mean_score_time', 
#                             'std_score_time']

#       # Move the important columns ahead
#       cv_results = cv_results[important_columns + sorted(list(set(cv_results.columns) - set(important_columns)))]

#       # Write cv_results file
#       cv_results.to_csv(path_or_buf=abspath_curr + name + '/' + acronym + '.csv', index=False)

#   #************************************************************************************************
#   # Test

#   # The list of [precision, recall, fscore, auc, best_estimator]
#   precision_recall_fscore_auc_best_estimator = []

#   for best_score, best_param, best_estimator in best_score_param_estimator_gs:
#       # Get the prediction
#       y_pred = best_estimator.predict(X_test)

#       # Get the precision, recall, fscore, support
#       precision, recall, fscore, support = precision_recall_fscore_support(y_test, y_pred)

#       # Get the auc
#       auc = roc_auc_score(y_test, y_pred)

#       # Update precision_recall_fscore_auc_best_estimator
#       precision_recall_fscore_auc_best_estimator.append([precision, recall, fscore, auc, best_estimator])

#   # Return precision_recall_fscore_best_estimator
#   return pd.DataFrame(precision_recall_fscore_auc_best_estimator, columns=['Precision', 'Recall', 'F1-score', 'AUC', 'Model'])


#%%


import pandas as pd
import numpy as np
import sys
# import os
# import matplotlib.pyplot as plt
from sklearn.base import BaseEstimator, RegressorMixin
# from sklearn.manifold import TSNE
# from sklearn.model_selection import PredefinedSplit
# from sklearn.model_selection import GridSearchCV
# from sklearn.model_selection import StratifiedKFold
# from sklearn.metrics import precision_recall_fscore_support
# from sklearn.metrics import roc_auc_score

#%%
# ### The normal equation
class LinearRegression_NE(BaseEstimator, RegressorMixin):
  """Linear regression implemented using the normal equation (Yuxiao Huang)"""

  def fit(self, X, y):
    """
    The fit function
    
    Parameters
    ----------
    X : the feature matrix
    y : the target vector
    """
    
    # Get the augmented feature matrix, [1, X]
    IX = np.hstack((np.ones((X.shape[0], 1)), X))

    # Get the optimal solution using the normal equation
    self.theta = np.linalg.pinv(IX).dot(y)
    # pinv() gives the Moore-Penrose pseudo-inverse of a matrix 
    # particularly useful when matrix A is not a sqaure, or A is singular.
    
    # Get the predicted target vector
    y_pred = self.net_input(IX)
                    
    # Get the loss (MSE)
    self.loss = ((y - y_pred) ** 2).sum() / IX.shape[0]
      
  def net_input(self, IX):
    """
    Get the predicted target vector
    
    Parameters
    ----------
    IX : The augmented feature matrix [1, X]
    
    Returns
    ----------
    The predicted target vector
    """
    
    return IX.dot(self.theta)

  def predict(self, X):
    """
    The predict function
    
    Parameters
    ----------
    X : the feature matrix
    
    Returns
    ----------
    The predicted value of the target
    """
    
    # Get the augmented feature matrix [1, X]
    IX = np.hstack((np.ones((X.shape[0], 1)), X))
    
    return self.net_input(IX)

#%%
# ### Batch gradient descent (BGD)
class LinearRegression_BGD(BaseEstimator, RegressorMixin):
  """Linear regression implemented using batch gradient descent and regularization (lasso, ridge and elastic net)"""
  
  def __init__(self, 
              max_iter=100, 
              eta=10 ** -2,
              penalty='l2',
              alpha=0.0001, 
              gamma=0.15,
              random_state=42):
    
    # The maximum number of epochs
    self.max_iter = max_iter
    
    # The learning rate
    self.eta = eta
    
    # The regularization term
    self.penalty=penalty
    
    # The regularization parameter
    self.alpha=alpha

    # The elastic net mixing parameter
    self.gamma=gamma

    # The random state
    self.random_state = random_state

  def fit(self, X_train, y_train, X_val=None, y_val=None):
    """
    The fit function
    
    Parameters
    ----------
    X_train : The training feature matrix
    y_train : The training target vector
    X_val : The validation feature matrix
    y_val : The validation target vector
    """
    
    # Get the augmented training feature matrix, [1, X_train]
    IX_train = np.hstack((np.ones((X_train.shape[0], 1)), X_train))
    
    # Get the random number generator
    self.rgen = np.random.RandomState(seed=self.random_state)
    
    # Initialize the parameters
    self.theta = self.rgen.normal(loc=0.0, scale=0.01, size=IX_train.shape[1])
    
    # Initialize the training and validation loss
    self.loss_train, self.loss_val = [], []
    
    # For each epoch
    for _ in range(self.max_iter):
      # Get the predicted target vector on the training data
      y_train_pred = self.net_input(IX_train)
      
      # Get the training error
      error_train = y_train - y_train_pred
                  
      # Get the training mse
      mse_train = (error_train ** 2).sum() / IX_train.shape[0]
      
      # Update the parameters
      # If no regularization
      if self.penalty == None:
          self.theta += self.eta * (2 / IX_train.shape[0] * IX_train.T.dot(error_train))
      # If lasso
      elif self.penalty == 'l1':
          self.theta += self.eta * (2 / IX_train.shape[0] * IX_train.T.dot(error_train) 
                                    - self.alpha * np.append([0], np.sign(self.theta[1:])))
      # If ridge
      elif self.penalty == 'l2':
          self.theta += self.eta * (2 / IX_train.shape[0] * IX_train.T.dot(error_train) 
                                    - self.alpha * np.append([0], self.theta[1:]))
      # If elastic net
      elif self.penalty == 'elasticnet':
          self.theta += self.eta * (2 / IX_train.shape[0] * IX_train.T.dot(error_train) 
                                    - self.alpha * self.gamma * np.append([0], np.sign(self.theta[1:])) 
                                    - self.alpha * (1 - self.gamma) * np.append([0], self.theta[1:]))
                      
      # Update the training loss
      self.loss_train.append(mse_train)
      
      # If the validation feature matrix and target vector are available
      if X_val is not None and y_val is not None:
        # Get the augmented validation feature matrix, [1, X_val]
        IX_val = np.hstack((np.ones((X_val.shape[0], 1)), X_val))
        
        # Get the predicted target vector on the validation data
        y_val_pred = self.net_input(IX_val)
        
        # Get the validation error
        error_val = y_val - y_val_pred
        
        # Get the validation mse
        mse_val = (error_val ** 2).sum() / IX_val.shape[0]
        
        # Update the validation loss
        self.loss_val.append(mse_val)

  def net_input(self, IX):
    """
    Get the predicted target vector
    
    Parameters
    ----------
    IX : The augmented feature matrix [1, X]
    
    Returns
    ----------
    The predicted target vector
    """
    
    return IX.dot(self.theta)

  def predict(self, X):
    """
    The predict function
    
    Parameters
    ----------
    X : the feature matrix
    
    Returns
    ----------
    The predicted value of the target
    """
    
    # Get the augmented feature matrix [1, X]
    IX = np.hstack((np.ones((X.shape[0], 1)), X))
    
    return self.net_input(IX)

#%%
# ### Stochastic gradient descent (SGD)
class LinearRegression_SGD(BaseEstimator, RegressorMixin):
  """Linear regression implemented using stochastic gradient descent and regularization (lasso, ridge and elastic net)"""
  
  def __init__(self, 
              max_iter=100,
              shuffle=True,
              eta=10 ** -2, 
              penalty='l2',
              alpha=0.0001, 
              gamma=0.15,
              random_state=42):
    
    # The maximum number of epochs
    self.max_iter = max_iter
    
    # Whether to shuffle samples in each epoch
    self.shuffle = shuffle
    
    # The learning rate
    self.eta = eta
    
    # The regularization term
    self.penalty=penalty
    
    # The regularization parameter
    self.alpha=alpha

    # The elastic net mixing parameter
    self.gamma=gamma
    
    # The random state
    self.random_state = random_state

  def fit(self, X_train, y_train, X_val=None, y_val=None):
    """
    The fit function
    
    Parameters
    ----------
    X_train : The training feature matrix
    y_train : The training target vector
    X_val : The validation feature matrix
    y_val : The validation target vector
    """
    
    # Get the augmented training feature matrix, [1, X_train]
    IX_train = np.hstack((np.ones((X_train.shape[0], 1)), X_train))
    
    # Get the indices of the augmented training feature matrix
    idxs_train = np.array(range(IX_train.shape[0]))
    
    # Get the random number generator
    self.rgen = np.random.RandomState(seed=self.random_state)
    
    # Initialize the parameters
    self.theta = self.rgen.normal(loc=0.0, scale=0.01, size=IX_train.shape[1])
    
    # Initialize the training and validation loss
    self.loss_train, self.loss_val = [], []
    
    # For each epoch
    for _ in range(self.max_iter):
      if self.shuffle is True:
        # Shuffle the indices
        self.rgen.shuffle(idxs_train)
          
      # Initialize the mse
      mse_train = 0
      
      # For each sample
      for i in idxs_train:                
        # Get the predicted target vector on the training data
        y_train_pred = self.net_input(IX_train[i, :])

        # Get the training error
        error_train = y_train[i] - y_train_pred

        # Get the training mse
        mse_train += (error_train ** 2) / IX_train.shape[0]

        # Update the parameters
        # If no regularization
        if self.penalty == None:
          self.theta += self.eta * (2 * IX_train[i, :].T.dot(error_train))
        # If lasso
        elif self.penalty == 'l1':
          self.theta += self.eta * (2 * IX_train[i, :].T.dot(error_train) 
                                    - self.alpha * np.append([0], np.sign(self.theta[1:])))
        # If ridge
        elif self.penalty == 'l2':
          self.theta += self.eta * (2 * IX_train[i, :].T.dot(error_train) 
                                    - self.alpha * np.append([0], self.theta[1:]))
        # If elastic net
        elif self.penalty == 'elasticnet':
          self.theta += self.eta * (2 * IX_train[i, :].T.dot(error_train) 
                                    - self.alpha * self.gamma * np.append([0], np.sign(self.theta[1:])) 
                                    - self.alpha * (1 - self.gamma) * np.append([0], self.theta[1:]))

      # Update the training loss
      self.loss_train.append(mse_train)

      # If the validation feature matrix and target vector are available
      if X_val is not None and y_val is not None:
        # Get the augmented validation feature matrix, [1, X_val]
        IX_val = np.hstack((np.ones((X_val.shape[0], 1)), X_val))

        # Get the predicted target vector on the validation data
        y_val_pred = self.net_input(IX_val)

        # Get the validation error
        error_val = y_val - y_val_pred

        # Get the validation mse
        mse_val = (error_val ** 2).sum() / IX_val.shape[0]

        # Update the validation loss
        self.loss_val.append(mse_val)
              
  def net_input(self, IX):
    """
    Get the predicted target vector
    
    Parameters
    ----------
    IX : The augmented feature matrix [1, X]
    
    Returns
    ----------
    The predicted target vector
    """
    
    return IX.dot(self.theta)

  def predict(self, X):
    """
    The predict function
    
    Parameters
    ----------
    X : the feature matrix
    
    Returns
    ----------
    The predicted value of the target
    """
    
    # Get the augmented feature matrix [1, X]
    IX = np.hstack((np.ones((X.shape[0], 1)), X))
    
    return self.net_input(IX)

#%%
# ### Mini-batch gradient descent (MBGD)
class LinearRegression_MBGD(BaseEstimator, RegressorMixin):
  """Linear regression implemented using mini-batch gradient descent and regularization (lasso, ridge and elastic net)"""
  
  def __init__(self,
                max_iter=100,
                shuffle=True,
                batch_size=32,
                eta=10 ** -2, 
                penalty='l2',
                alpha=0.0001, 
                gamma=0.15,
                random_state=42):
      
    # The maximum number of epochs
    self.max_iter = max_iter
    
    # Whether to shuffle samples in each epoch
    self.shuffle = shuffle
    
    # The size of minibatches for stochastic optimizers
    self.batch_size = batch_size
    
    # The learning rate
    self.eta = eta
    
    # The regularization term
    self.penalty=penalty
    
    # The regularization parameter
    self.alpha=alpha

    # The elastic net mixing parameter
    self.gamma=gamma
    
    # The random state
    self.random_state = random_state

  def fit(self, X_train, y_train, X_val=None, y_val=None):
    """
    The fit function
    
    Parameters
    ----------
    X_train : The training feature matrix
    y_train : The training target vector
    X_val : The validation feature matrix
    y_val : The validation target vector
    """
    
    # Get the augmented training feature matrix, [1, X_train]
    IX_train = np.hstack((np.ones((X_train.shape[0], 1)), X_train))
    
    # Get the random number generator
    self.rgen = np.random.RandomState(seed=self.random_state)
    
    # Initialize the parameters
    self.theta = self.rgen.normal(loc=0.0, scale=0.01, size=IX_train.shape[1])
    
    # Initialize the training and validation loss
    self.loss_train, self.loss_val = [], []
    
    # For each epoch
    for _ in range(self.max_iter):
      # Get the indices of the training data
      idxs_train = np.array(range(IX_train.shape[0]))
      
      # Get the minibatches of the training data
      mbs = self.get_minibatches(idxs_train)

      # Initialize the training mse
      mse_train = 0

      # For each minibatch
      for mb in mbs:   
        # Get the augmented training feature matrix and target vector
        IX_train_mb, y_train_mb = IX_train[mb,:], y_train[mb]

        # Get the predicted target vector on the training data
        y_train_mb_pred = self.net_input(IX_train_mb)

        # Get the training error
        error_train = y_train_mb - y_train_mb_pred

        # Get the training mse
        mse_train += (error_train ** 2).sum() / IX_train.shape[0]
        
        # Update the parameters
        # If no regularization
        if self.penalty == None:
            self.theta += self.eta * (2 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train))
        # If lasso
        elif self.penalty == 'l1':
            self.theta += self.eta * (2 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * np.append([0], np.sign(self.theta[1:])))
        # If ridge
        elif self.penalty == 'l2':
            self.theta += self.eta * (2 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * np.append([0], self.theta[1:]))
        # If elastic net
        elif self.penalty == 'elasticnet':
            self.theta += self.eta * (2 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * self.gamma * np.append([0], np.sign(self.theta[1:])) 
                                      - self.alpha * (1 - self.gamma) * np.append([0], self.theta[1:]))

      # Update the training loss
      self.loss_train.append(mse_train)
      
      # If the validation feature matrix and target vector are available
      if X_val is not None and y_val is not None:
        # Get the augmented validation feature matrix, [1, X_val]
        IX_val = np.hstack((np.ones((X_val.shape[0], 1)), X_val))
        
        # Get the predicted target vector on the validation data
        y_val_pred = self.net_input(IX_val)
        
        # Get the validation error
        error_val = y_val - y_val_pred
        
        # Get the validation mse
        mse_val = (error_val ** 2).sum() / IX_val.shape[0]
        
        # Update the validation loss
        self.loss_val.append(mse_val)

  def get_minibatches(self, idxs):
    """
    Get the minibatches
    
    Parameters
    ----------
    idxs : The indices of the data
    
    Returns
    ----------
    The minibatches
    """
    
    # Initialize the minibatches
    mbs = []
    
    if self.shuffle is True:
      # Shuffle the indices
      self.rgen.shuffle(idxs)
            
    # Get the number of minibatches
    n_batch = len(idxs) // self.batch_size
    
    # For each minibatch
    for i in range(n_batch):
      # Get the first and last index (exclusive) of the minibatch
      first_idx = i * self.batch_size
      last_idx = min((i + 1) * self.batch_size, len(idxs))
                              
      # Get the minibatch
      mb = idxs[first_idx : last_idx]
      
      # Update the minibatches
      mbs.append(mb)

    return mbs

  def net_input(self, IX):
    """
    Get the predicted target vector
    
    Parameters
    ----------
    IX : The augmented feature matrix [1, X]
    
    Returns
    ----------
    The predicted target vector
    """
    
    return IX.dot(self.theta)

  def predict(self, X):
    """
    The predict function
    
    Parameters
    ----------
    X : the feature matrix
    
    Returns
    ----------
    The predicted value of the target
    """
    
    # Get the augmented feature matrix [1, X]
    IX = np.hstack((np.ones((X.shape[0], 1)), X))
    
    return self.net_input(IX)
      
#%%
# ### Mini-batch gradient descent (MBGD)
class LogisticRegression_MBGD(BaseEstimator, RegressorMixin):
  """Logistic regression implemented using mini-batch gradient descent and regularization (lasso, ridge and elastic net)"""
  
  def __init__(self,
                max_iter=100,
                shuffle=True,
                batch_size=32,
                eta=10 ** -2, 
                penalty='l2',
                alpha=1, 
                gamma=0.5,
                random_state=42):
      
      # The maximum number of epochs
      self.max_iter = max_iter
      
      # Whether to shuffle samples in each epoch
      self.shuffle = shuffle
      
      # The size of minibatches for stochastic optimizers
      self.batch_size = batch_size
      
      # The learning rate
      self.eta = eta
      
      # The regularization term
      self.penalty=penalty
      
      # The regularization parameter
      self.alpha=alpha

      # The elastic net mixing parameter
      self.gamma=gamma
      
      # The random state
      self.random_state = random_state

  def fit(self, X_train, y_train, X_val=None, y_val=None):
    """
    The fit function
    
    Parameters
    ----------
    X_train : The training feature matrix
    y_train : The training target vector
    X_val : The validation feature matrix
    y_val : The validation target vector
    """
    
    # Get the augmented training feature matrix, [1, X_train]
    IX_train = np.hstack((np.ones((X_train.shape[0], 1)), X_train))
    
    # Get the random number generator
    self.rgen = np.random.RandomState(seed=self.random_state)
    
    # Get the unique classes of the target
    self.classes = np.unique(y_train)
    
    # Get the number of unique classes of the target
    self.n_classes = len(self.classes)
    
    # If binary classification
    if self.n_classes == 2:
      # Make a copy of y_train
      Y_train = np.copy(y_train)       
    # If multi-class classification
    else:
      # Get the one-hot-encoded training target matrix
      Y_train = pd.get_dummies(y_train).values
    
    # Initialize the parameters
    # If binary classification
    if self.n_classes == 2:
      self.theta = self.rgen.normal(loc=0.0, scale=0.01, size=IX_train.shape[1])
    # If multi-class classification
    else:
      self.theta = self.rgen.normal(loc=0.0, scale=0.01, size=(IX_train.shape[1], self.n_classes))            
    
    # Initialize the training and validation loss
    self.loss_train, self.loss_val = [], []
    
    # For each epoch
    for _ in range(self.max_iter):
      # Get the indices of the training data
      idxs_train = np.array(range(IX_train.shape[0]))
      
      # Get the minibatches of the training data
      mbs = self.get_minibatches(idxs_train)

      # Initialize the training mse
      mse_train = 0

      # For each minibatch
      for mb in mbs:   
        # Get the augmented training feature matrix and target matrix
        # If binary classification
        if self.n_classes == 2:
          IX_train_mb, Y_train_mb = IX_train[mb,:], Y_train[mb]       
        # If multi-class classification
        else:
          IX_train_mb, Y_train_mb = IX_train[mb,:], Y_train[mb,:]       
                                  
        # Get the net input matrix
        N_train_mb = self.net_input(IX_train_mb)
        
        # Get the probability matrix
        P_train_mb = self.activation(N_train_mb)
                                  
        # Get the training error
        error_train = Y_train_mb - P_train_mb

        # Get the training mse
        mse_train += (error_train ** 2).sum() / IX_train.shape[0]
        
        # Update the parameters
        # If no regularization
        if self.penalty == None:
          self.theta += self.eta / IX_train_mb.shape[0] * (IX_train_mb.T.dot(error_train))
        # If lasso
        elif self.penalty == 'l1':
          # If binary classification
          if self.n_classes == 2:
              self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                        - self.alpha * np.append([0], np.sign(self.theta[1:])))
          # If multi-class classification
          else:
              self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                        - self.alpha * np.append(np.zeros((1, self.theta.shape[1])), np.sign(self.theta[1:,:]), axis=0))                        
        # If ridge
        elif self.penalty == 'l2':
          # If binary classification
          if self.n_classes == 2:
              self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                        - self.alpha * np.append([0], self.theta[1:]))
          # If multi-class classification
          else:
              self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                        - self.alpha * np.append(np.zeros((1, self.theta.shape[1])), self.theta[1:,:], axis=0))                        
        # If elastic net
        elif self.penalty == 'elasticnet':
          # If binary classification
          if self.n_classes == 2:
              self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                        - self.alpha * self.gamma * np.append([0], np.sign(self.theta[1:])) 
                                        - self.alpha * (1 - self.gamma) * np.append([0], self.theta[1:]))
          # If multi-class classification
          else:
              self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                        - self.alpha * self.gamma * np.append(np.zeros((1, self.theta.shape[1])), np.sign(self.theta[1:,:]), axis=0) 
                                        - self.alpha * (1 - self.gamma) * np.append(np.zeros((1, self.theta.shape[1])), self.theta[1:,:], axis=0))
                                        
      # Update the training loss
      self.loss_train.append(mse_train)
      
      # If the validation feature matrix and target vector are available
      if X_val is not None and y_val is not None:
        # Get the augmented validation feature matrix, [1, X_val]
        IX_val = np.hstack((np.ones((X_val.shape[0], 1)), X_val))

        # If binary classification
        if self.n_classes == 2:
          # Make a copy of y_val
          Y_val = np.copy(y_val)            
        # If multi-class classification
        else:
          # Get the one-hot-encoded validation target matrix
          Y_val = pd.get_dummies(y_val).values
        
        # Get the net input matrix on the validation data
        N_val = self.net_input(IX_val)
        
        # Get the probability matrix on the validation data
        P_val = self.activation(N_val)
                                  
        # Get the validation error
        error_val = Y_val - P_val
        
        # Get the validation mse
        mse_val = (error_val ** 2).sum() / IX_val.shape[0]
        
        # Update the validation loss
        self.loss_val.append(mse_val)

  def get_minibatches(self, idxs):
    """
    Get the minibatches
    
    Parameters
    ----------
    idxs : The indices of the data
    
    Returns
    ----------
    The minibatches
    """
    
    # Initialize the minibatches
    mbs = []
    
    if self.shuffle is True:
      # Shuffle the indices
      self.rgen.shuffle(idxs)
            
    # Get the number of minibatches
    n_batch = len(idxs) // self.batch_size
    
    # For each minibatch
    for i in range(n_batch):
      # Get the first and last index (exclusive) of the minibatch
      first_idx = i * self.batch_size
      last_idx = min((i + 1) * self.batch_size, len(idxs))
                              
      # Get the minibatch
      mb = idxs[first_idx : last_idx]
      
      # Update the minibatches
      mbs.append(mb)

    return mbs

  def net_input(self, IX):
    """
    Get the net input matrix
    
    Parameters
    ----------
    IX : The augmented feature matrix [1, X]
    
    Returns
    ----------
    The net input matrix
    """
    
    return IX.dot(self.theta)
  
  def activation(self, net_input):
    """
    Get the probability (sigmoid or softmax) matrix
    
    Parameters
    ----------
    net_input : The net input
    
    Returns
    ----------
    The probability (sigmoid or softmax) matrix
    """
    
    # If binary classification
    if self.n_classes == 2:
      # Get the exponent of the negative net input
      neg_net_input_exp = np.exp(-np.clip(net_input, -250, 250))
      
      # Return the sigmoid matrix
      return 1. / (1. + neg_net_input_exp)           
    # If multi-class classification
    else:
      # Get the exponent of the net input
      net_input_exp = np.exp(net_input - np.max(net_input, axis=1).reshape(-1, 1))

      # Return the softmax matrix
      return net_input_exp / np.sum(net_input_exp, axis=1).reshape(-1, 1)
      
  def predict_proba(self, X):
    """
    The predict probability function
    
    Parameters
    ----------
    X : the feature matrix
    
    Returns
    ----------
    The probability (sigmoid or softmax) matrix
    """
    
    # Get the augmented feature matrix [1, X]
    IX = np.hstack((np.ones((X.shape[0], 1)), X))
        
    # Get the net_input matrix
    N = self.net_input(IX)

    return self.activation(N)
  
  def predict(self, X):
    """
    The predict class function
    
    Parameters
    ----------
    X : the feature matrix
    
    Returns
    ----------
    The predicted class vector
    """
    
    # If binary classification
    if self.n_classes == 2:
      return (self.predict_proba(X) >= 0.5) * 1         
    # If multi-class classification
    else:
      return np.argmax(self.predict_proba(X), axis=1)

#%%
# ## Shallow neural networks 
# ### Single-layer perceptron
class SingleLayerPerceptron_MBGD(BaseEstimator, RegressorMixin):
  """Single-layer perceptron implemented using mini-batch gradient descent and regularization (lasso, ridge and elastic net)"""
  
  def __init__(self,
              max_iter=100,
              shuffle=True,
              batch_size=32,
              eta=10 ** -2, 
              penalty='l2',
              alpha=1, 
              gamma=0.5,
              random_state=42):
    
    # The maximum number of epochs
    self.max_iter = max_iter
    
    # Whether to shuffle samples in each epoch
    self.shuffle = shuffle
    
    # The size of minibatches for stochastic optimizers
    self.batch_size = batch_size
    
    # The learning rate
    self.eta = eta
    
    # The regularization term
    self.penalty=penalty
    
    # The regularization parameter
    self.alpha=alpha

    # The elastic net mixing parameter
    self.gamma=gamma
    
    # The random state
    self.random_state = random_state

  def fit(self, X_train, y_train, X_val=None, y_val=None):
    """
    The fit function
    
    Parameters
    ----------
    X_train : The training feature matrix
    y_train : The training target vector
    X_val : The validation feature matrix
    y_val : The validation target vector
    """
    
    # Get the augmented training feature matrix, [1, X_train]
    IX_train = np.hstack((np.ones((X_train.shape[0], 1)), X_train))
    
    # Get the random number generator
    self.rgen = np.random.RandomState(seed=self.random_state)
    
    # Get the unique classes of the target
    self.classes = np.unique(y_train)
    
    # Get the number of unique classes of the target
    self.n_classes = len(self.classes)
    
    # If binary classification
    if self.n_classes == 2:
      # Make a copy of y_train
      Y_train = np.copy(y_train)       
    # If multi-class classification
    else:
      # Get the one-hot-encoded training target matrix
      Y_train = pd.get_dummies(y_train).values
    
    # Initialize the parameters
    # If binary classification
    if self.n_classes == 2:
      self.theta = self.rgen.normal(loc=0.0, scale=0.01, size=IX_train.shape[1])
    # If multi-class classification
    else:
      self.theta = self.rgen.normal(loc=0.0, scale=0.01, size=(IX_train.shape[1], self.n_classes))            
    
    # Initialize the training and validation loss
    self.loss_train, self.loss_val = [], []
    
    # For each epoch
    for _ in range(self.max_iter):
      # Get the indices of the training data
      idxs_train = np.array(range(IX_train.shape[0]))
      
      # Get the minibatches of the training data
      mbs = self.get_minibatches(idxs_train)

      # Initialize the training mse
      mse_train = 0

      # For each minibatch
      for mb in mbs:   
        # Get the augmented training feature matrix and target matrix
        # If binary classification
        if self.n_classes == 2:
          IX_train_mb, Y_train_mb = IX_train[mb,:], Y_train[mb]       
        # If multi-class classification
        else:
          IX_train_mb, Y_train_mb = IX_train[mb,:], Y_train[mb,:]       
                                  
        # Get the net input matrix
        N_train_mb = self.net_input(IX_train_mb)
        
        # Get the output matrix
        A_train_mb = self.activation(N_train_mb)
                                  
        # Get the training error
        error_train = Y_train_mb - A_train_mb

        # Get the training mse
        mse_train += (error_train ** 2).sum() / IX_train.shape[0]
        
        # Update the parameters
        # If no regularization
        if self.penalty == None:
          self.theta += self.eta / IX_train_mb.shape[0] * (IX_train_mb.T.dot(error_train))
        # If lasso
        elif self.penalty == 'l1':
          # If binary classification
          if self.n_classes == 2:
            self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * np.append([0], np.sign(self.theta[1:])))
          # If multi-class classification
          else:
            self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * np.append(np.zeros((1, self.theta.shape[1])), np.sign(self.theta[1:,:]), axis=0))                        
        # If ridge
        elif self.penalty == 'l2':
          # If binary classification
          if self.n_classes == 2:
            self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * np.append([0], self.theta[1:]))
          # If multi-class classification
          else:
            self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * np.append(np.zeros((1, self.theta.shape[1])), self.theta[1:,:], axis=0))                        
        # If elastic net
        elif self.penalty == 'elasticnet':
          # If binary classification
          if self.n_classes == 2:
            self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * self.gamma * np.append([0], np.sign(self.theta[1:])) 
                                      - self.alpha * (1 - self.gamma) * np.append([0], self.theta[1:]))
          # If multi-class classification
          else:
            self.theta += self.eta * (1 / IX_train_mb.shape[0] * IX_train_mb.T.dot(error_train) 
                                      - self.alpha * self.gamma * np.append(np.zeros((1, self.theta.shape[1])), np.sign(self.theta[1:,:]), axis=0) 
                                      - self.alpha * (1 - self.gamma) * np.append(np.zeros((1, self.theta.shape[1])), self.theta[1:,:], axis=0))
                                          
      # Update the training loss
      self.loss_train.append(mse_train)
      
      # If the validation feature matrix and target vector are available
      if X_val is not None and y_val is not None:
        # Get the augmented validation feature matrix, [1, X_val]
        IX_val = np.hstack((np.ones((X_val.shape[0], 1)), X_val))

        # If binary classification
        if self.n_classes == 2:
          # Make a copy of y_val
          Y_val = np.copy(y_val)            
        # If multi-class classification
        else:
          # Get the one-hot-encoded validation target matrix
          Y_val = pd.get_dummies(y_val).values
        
        # Get the net input matrix on the validation data
        N_val = self.net_input(IX_val)
        
        # Get the output matrix on the validation data
        A_val = self.activation(N_val)
                                  
        # Get the validation error
        error_val = Y_val - A_val
        
        # Get the validation mse
        mse_val = (error_val ** 2).sum() / IX_val.shape[0]
        
        # Update the validation loss
        self.loss_val.append(mse_val)

  def get_minibatches(self, idxs):
    """
    Get the minibatches
    
    Parameters
    ----------
    idxs : The indices of the data
    
    Returns
    ----------
    The minibatches
    """
    
    # Initialize the minibatches
    mbs = []
    
    if self.shuffle is True:
      # Shuffle the indices
      self.rgen.shuffle(idxs)
            
    # Get the number of minibatches
    n_batch = len(idxs) // self.batch_size
    
    # For each minibatch
    for i in range(n_batch):
      # Get the first and last index (exclusive) of the minibatch
      first_idx = i * self.batch_size
      last_idx = min((i + 1) * self.batch_size, len(idxs))
                              
      # Get the minibatch
      mb = idxs[first_idx : last_idx]
      
      # Update the minibatches
      mbs.append(mb)

    return mbs

  def net_input(self, IX):
    """
    Get the net input matrix
    
    Parameters
    ----------
    IX : The augmented feature matrix [1, X]
    
    Returns
    ----------
    The net input matrix
    """
    
    return IX.dot(self.theta)
  
  def activation(self, net_input):
    """
    Get the probability (sigmoid or softmax) matrix
    
    Parameters
    ----------
    net_input : The net input
    
    Returns
    ----------
    The probability (sigmoid or softmax) matrix
    """
    
    # If binary classification
    if self.n_classes == 2:
        return (net_input >= 0) * 1          
    # If multi-class classification
    else:
      # Get the exponent of the net input
      net_input_exp = np.exp(net_input - np.max(net_input, axis=1).reshape(-1, 1))

      # Return the softmax matrix
      return net_input_exp / np.sum(net_input_exp, axis=1).reshape(-1, 1)
      
  def predict_proba(self, X):
      """
      The predict probability function
      
      Parameters
      ----------
      X : the feature matrix
      
      Returns
      ----------
      The probability (sigmoid or softmax) matrix
      """
      
      # Get the augmented feature matrix [1, X]
      IX = np.hstack((np.ones((X.shape[0], 1)), X))
          
      # Get the net_input matrix
      N = self.net_input(IX)

      return self.activation(N)
  
  def predict(self, X):
      """
      The predict class function
      
      Parameters
      ----------
      X : the feature matrix
      
      Returns
      ----------
      The predicted class vector
      """
      
      # If binary classification
      if self.n_classes == 2:
        return (self.predict_proba(X) >= 0.5) * 1         
      # If multi-class classification
      else:
        return np.argmax(self.predict_proba(X), axis=1)
          
#%%
# ### Multiple-layer perceptron
class MultiLayerPerceptron_MBGD(BaseEstimator, RegressorMixin):
  """Multi-layer perceptron implemented using mini-batch gradient descent and regularization (lasso, ridge and elastic net)"""
      
  def __init__(self, 
              hidden_layer_sizes=[100], 
              activation='relu', 
              batch_size=32,
              learning_rate_init=0.01, 
              max_iter=100, 
              shuffle=True, 
              penalty='l2',
              alpha=1, 
              gamma=0.5,
              random_state=42):
    
    # The number of perceptrons on each hidden layer
    self.hidden_layer_sizes = hidden_layer_sizes
    
    # The Activation function
    self.activation = activation
    
    # The size of minibatches for stochastic optimizers
    self.batch_size = batch_size
    
    # The initial learning rate
    self.learning_rate_init = learning_rate_init

    # The maximum number of epochs
    self.max_iter = max_iter
    
    # Whether to shuffle samples in each epoch
    self.shuffle = shuffle
    
    # The regularization term
    self.penalty=penalty
    
    # The regularization parameter
    self.alpha=alpha

    # The elastic net mixing parameter
    self.gamma=gamma
    
    # The random state
    self.random_state = random_state

  def fit(self, X, y):
    """
    The fit function
    
    Parameters
    ----------
    X : the feature matrix
    y : the target vector
    """
    
    # Get the one-hot-encoded target matrix
    Y = pd.get_dummies(y).values
    
    # Initialize the object variables
    self.fit_init(X, Y)
            
    # For each epoch
    for epoch in range(self.max_iter):
      # Get the indices of the training data
      idxs = np.array(range(X.shape[0]))
      
      # Get the minibatches of the training data
      mbs = self.get_mbes(idxs)
      
      # For each minibatch
      for mb in mbs:   
        # Get the feature matrix
        X_mb = X[mb,:]
                        
        # Get the target matrix
        Y_mb = Y[mb,:]
                        
        # Update the weights and biases using mini-batch gradient descent
        self.mini_batch_gradient_descent(X_mb.reshape(self.batch_size, -1), Y_mb.reshape(self.batch_size, -1))
        
        # Update the cost
        self.costs[epoch] += np.sum((Y_mb.reshape(self.batch_size, -1) - self.activations[-1]) ** 2) / self.m
              
  def fit_init(self, X, Y):
    """
    Initialize the object variables
    
    Parameters
    ----------
    X : the feature matrix
    y : the target matrix
    """
    
    # Initialize the number of samples and featurs
    self.m, self.n = X.shape
    
    # Initialize the number of unique class labels
    self.classes = np.unique(Y)
    
    # Initialize the number of perceptrons on each layer
    self.layer_sizes = [self.n] + self.hidden_layer_sizes + [len(self.classes)]
    
    # Initialize the cost
    self.costs = np.zeros(self.max_iter)
    
    # Initialize the random number generator
    self.rgen = np.random.RandomState(seed=self.random_state)
    
    # Initialize the weights
    self.W = [0] + [self.rgen.normal(loc=0.0, 
                                      scale=2 / np.sqrt( self.layer_sizes[i - 1] + self.layer_sizes[i]), 
                                      size=(self.layer_sizes[i - 1], self.layer_sizes[i]))
                    for i in range(1, len(self.layer_sizes))]
    
    # Initialize the biases
    self.b = [0] + [np.zeros((1, self.layer_sizes[i]))
                    for i in range(1, len(self.layer_sizes))]
    
    # Initialize the net inputs
    self.net_inputs = [0] * (len(self.layer_sizes))

    # Initialize the activations
    self.activations = [0] * (len(self.layer_sizes))
    
    # Initialize the sensitivities
    self.sensitivities = [0] * (len(self.layer_sizes))
    
    # Initialize the derivatives
    self.derivatives = [0] * (len(self.layer_sizes))
      
  def get_mbes(self, idxs):
    """
    Get the minibatches
    
    Parameters
    ----------
    idxs : The indices of the data
    
    Returns
    ----------
    The minibatches
    """
    
    # Initialize the minibatches
    mbs = []
    
    if self.shuffle is True:
      # Shuffle the indices
      self.rgen.shuffle(idxs)
            
    # Get the number of minibatches
    n_batch = len(idxs) // self.batch_size
    
    # For each minibatch
    for i in range(n_batch):
      # Get the first and last index (exclusive) of the minibatch
      first_idx = i * self.batch_size
      last_idx = min((i + 1) * self.batch_size, len(idxs))
                              
      # Get the minibatch
      mb = idxs[first_idx : last_idx]
      
      # Update the minibatches
      mbs.append(mb)

    return mbs
              
  def mini_batch_gradient_descent(self, X, Y):
    """
    Update the weights and biases using mini-batch gradient descent
    
    Parameters
    ----------
    X : the feature matrix
    Y : the target matrix
    """
    
    # Get the activation in the first layer
    self.activations[0] = X

    # Propagate the net input and activation forward through the network
    for i in range(1, len(self.layer_sizes)):
      # Get the net input on layer i
      self.net_inputs[i] = self.get_net_input(i)  
      
      # Get the activation on layer i
      self.activations[i] = self.get_activation(i)
                            
    # Get the sensitivity in the last layer
    self.sensitivities[-1] = - 2 / X.shape[0] * np.array([np.matmul((Y - self.activations[-1])[k, :],
                                                                    self.get_derivative(len(self.layer_sizes) - 1)[k, :, :]) 
                                                          for k in range(X.shape[0])])
                    
    # Propagate the sensitivites backward through the network
    for i in range(len(self.layer_sizes) - 2, 0, -1):
      # Get the derivative on layer i
      self.derivatives[i] = self.get_derivative(i)
                  
      # Get the sensitivity on layer i
      self.sensitivities[i] = self.get_sensitivity(i)
                    
    # Update the weights and biases using gradient descent
    for i in range(1, len(self.layer_sizes)):
      # Update the biases
      self.b[i] -= self.learning_rate_init * self.sensitivities[i].sum(axis=0).reshape(1, -1)
      
      # Update the weights
      self.W[i] -= self.learning_rate_init * (np.matmul(self.activations[i - 1].T, self.sensitivities[i]))
                                              
      # If lasso
      if self.penalty == 'l1':
        self.W[i] -= self.learning_rate_init * self.alpha * np.sign(self.W[i])                        
      # If ridge
      elif self.penalty == 'l2':
        self.W[i] -= self.learning_rate_init * self.alpha * self.W[i]                                              
      # If elastic net
      elif self.penalty == 'elasticnet':
        self.W[i] -= self.learning_rate_init * self.alpha * (self.gamma * np.sign(self.W[i]) + (1 - self.gamma) * self.W[i])                                              
          
  def get_net_input(self, i):
    """
    Get the net input on layer i
    
    Parameters
    ----------
    i : the ith layer
    
    Returns
    ----------
    The net input on layer i
    """
    
    return self.b[i] + np.matmul(self.activations[i - 1], self.W[i]) 

  def get_activation(self, i):
    """
    Get the activation on layer i
    
    Parameters
    ----------
    i : the ith layer
    
    Returns
    ----------
    The activation on layer i    
    """ 

    if self.activation == 'identity':
      return self.net_inputs[i]
    elif self.activation == 'logistic':
      return 1 / (1 + np.exp(-self.net_inputs[i]))
    elif self.activation == 'tanh':
      e_z = np.exp(self.net_inputs[i])
      e_neg_z = np.exp(-self.net_inputs[i])
      return (e_z - e_neg_z) / (e_z + e_neg_z)
    elif self.activation == 'relu':
      return np.clip(self.net_inputs[i], 0, None)
    else:
      print("Activation undefined!")
      sys.exit(1)
  
  def get_derivative(self, i):
    """
    Get the derivative on layer i
    
    Parameters
    ----------
    i : the ith layer
    
    Returns
    ----------
    The derivative on layer i    
    """ 
    
    if self.activation == 'identity':
      return np.array([np.identity(self.activations[i].shape[1]) 
                        for j in range(self.activations[i].shape[0])])
    elif self.activation == 'logistic':        
      return np.array([np.diag((self.activations[i] * 1 - self.activations[i])[j, :]) 
                        for j in range(self.activations[i].shape[0])])
    elif self.activation == 'tanh':
      return np.array([np.diag((1 - self.activations[i] ** 2)[j, :]) 
                        for j in range(self.activations[i].shape[0])])
    elif self.activation == 'relu':         
      return np.array([np.diag(np.where(self.net_inputs[i][j, :] >= 0, 1, 0).reshape(-1)) 
                        for j in range(self.net_inputs[i].shape[0])])
    else:
      print("Activation undefined!")
      sys.exit(1)
      
  def get_sensitivity(self, i):
    """
    Get the sensitivity on layer i
    
    Parameters
    ----------
    i : the ith layer
    
    Returns
    ----------
    The sensitivity on layer i    
    """ 
    
    # Get matrix multiplication
    M = np.matmul(self.sensitivities[i + 1], self.W[i + 1].T)
    
    return np.array([np.matmul(M[k, :], self.derivatives[i][k, :, :]) 
                      for k in range(M.shape[0])])
      
  def predict(self, X):
    """
    The predict function
    
    Parameters
    ----------
    X : the feature matrix
    
    Returns
    ----------
    The predicted class labels of the target
    """

    # Initialize the net inputs
    self.net_inputs = [0] * (len(self.layer_sizes))
    
    # Initialize the activations
    self.activations = [0] * (len(self.layer_sizes))
    # Initialize the activation in the first layer
    self.activations[0] = X

    # Propagate the input forward through the network
    for i in range(1, len(self.layer_sizes)):
      # Get the net input on layer i
      self.net_inputs[i] = self.get_net_input(i)  
      
      # Get the activation on layer i
      self.activations[i] = self.get_activation(i)
                            
    return np.argmax(self.activations[-1], axis=1)
      

#%%




# __all__ = ('api','dbcon','dfchk','boundary_plots')
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import mysql.connector
import tensorflow as tf
from mysql.connector import Error
import requests
import os
import sys
import matplotlib.pyplot as plt
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.manifold import TSNE
from sklearn.model_selection import PredefinedSplit
# from sklearn.model_selection import GridSearchCV
# from sklearn.model_selection import StratifiedKFold
# from sklearn.metrics import precision_recall_fscore_support
# from sklearn.metrics import roc_auc_score

# from . import api
from .api import * 
# from . import helper
from .learn.helper import * 
# from . import grads
from .learn.grads import * 
# from . import tf_utils
from .learn.tf_utils import * 
# from .learn import boundary_plots
from .learn.boundary_plots import * 



# __all__ = ('dfchk')
import tensorflow as tf

#%%
# from https://github.com/yuxiaohuang/teaching
def normalize(data, label):
  """
  Normalize the data

  Parameters
  ----------
  data: the data
  label: the label
  
  Returns
  ----------
  The normalized data and label
  """

  # Normalize the data
  data_normalized = tf.cast(data, tf.float64) / 255

  return data_normalized, label
  

#%%
def resize(data, label):
  """
  Resize the data into the default input size of the pretrained model

  Parameters
  ----------
  data: the data
  label: the label
  
  Returns
  ----------
  The resized data and label
  """

  # Resize the data into the default input size of the pretrained model
  data_resized = tf.image.resize(data, input_size)

  return data_resized, label

#%%
def preprocess_pretrain(data, label):
  """
  Preprocess the data using pretrained model

  Parameters
  ----------
  data: the data
  label: the label
  
  Returns
  ----------
  The preprocessed data using pretrained model
  """

  # Preprocess the data
  data_preprocessed = preprocess_input(data)

  return data_preprocessed, label

#%%
